/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Gremlins.h"

#include "CGremlins.h"
#include "CGremlinsStubs.h"

#include "AppPreferences.h"		// gPrefs
#include "Byteswapping.h"		// Canonical
#include "CPU_REG.h"			// SPCFLAG_SAVE_STATE
#include "ErrorHandling.h"		// Errors::ReportMessage
#include "Logging.h"			// OpenLogFile, CloseLogFile
#include "Miscellaneous.h"		// ResetCalibrationInfo
#include "Platform.h"			// Platform::GetMilliseconds
#include "SessionFile.h"		// SessionFile
#include "Strings.r.h"			// kStr_ values
#include "UAE_Utils.h"			// uae_memcpy



// -----------------------------
   #pragma mark *** Globals
// -----------------------------

static Gremlins			gGremlins;
static bool				gGremlinsOn;

static unsigned long	gGremlinStartTime;
static unsigned long	gGremlinStopTime;


// -----------------------------
   #pragma mark -
   #pragma mark *** Private Functions
// -----------------------------


static PPStream&	operator >> (PPStream&, AppInfo&);
static PPStream&	operator << (PPStream&, const AppInfo&);

static PPStream&	operator >> (PPStream&, AppPreferences::GremlinInfo&);
static PPStream&	operator << (PPStream&, const AppPreferences::GremlinInfo&);

static PPStream&	operator >> (PPStream&, AppInfoList&);
static PPStream&	operator << (PPStream&, const AppInfoList&);


// -----------------------------
   #pragma mark -
   #pragma mark *** Public Functions
// -----------------------------


// ---------------------------------------------------------------------------
//		 Gremlins_Initialize
// ---------------------------------------------------------------------------

void Gremlins_Initialize (void)
{
}


// ---------------------------------------------------------------------------
//		 Gremlins_Reset
// ---------------------------------------------------------------------------

void Gremlins_Reset (void)
{
	Gremlins_Stop ();
}


// ---------------------------------------------------------------------------
//		 Gremlins_Save
// ---------------------------------------------------------------------------

void Gremlins_Save (SessionFile& f)
{
	if (::Gremlins_IsOn ())
		gGremlinStopTime = Platform::GetMilliseconds ();

	const long	kCurrentVersion = 2;

	Chunk		chunk;
	ChunkStream	s (chunk);

	s << kCurrentVersion;

	s << gGremlins.keyProbabilitiesSum;
	s << gGremlins.lastPoint.v;
	s << gGremlins.lastPoint.h;
	s << gGremlins.lastPenDown;
	s << gGremlins.number;
	s << gGremlins.counter;
	s << gGremlins.until;
	s << gGremlins.saveUntil;
	s << gGremlins.inited;
	s << gGremlins.catchUp;
	s << gGremlins.needPenUp;
	s << gGremlins.charsToType;

	s << gGremlinsOn;
	s << gGremlinStartTime;
	s << gGremlinStopTime;

	s << gGremlinNext;

	AppPreferences::GremlinInfo&	info = gPrefs->GetGremlinInfo ();

	s << info;

	f.WriteGremlinInfo (chunk);
}


// ---------------------------------------------------------------------------
//		 Gremlins_Load
// ---------------------------------------------------------------------------

void Gremlins_Load (SessionFile& f)
{
	Chunk	chunk;
	if (f.ReadGremlinInfo (chunk))
	{
		long		version;
		ChunkStream	s (chunk);

		s >> version;

		if (version >= 1)
		{
			s >> gGremlins.keyProbabilitiesSum;
			s >> gGremlins.lastPoint.v;
			s >> gGremlins.lastPoint.h;
			s >> gGremlins.lastPenDown;
			s >> gGremlins.number;
			s >> gGremlins.counter;
			s >> gGremlins.until;
			s >> gGremlins.saveUntil;
			s >> gGremlins.inited;
			s >> gGremlins.catchUp;
			s >> gGremlins.needPenUp;
			s >> gGremlins.charsToType;

			s >> gGremlinsOn;
			s >> gGremlinStartTime;
			s >> gGremlinStopTime;

			s >> gGremlinNext;

			// Patch up the start and stop times.

			uae_s32	delta = gGremlinStopTime - gGremlinStartTime;
			gGremlinStopTime = Platform::GetMilliseconds ();
			gGremlinStartTime = gGremlinStopTime - delta;
		}

		if (version >= 2)
		{
			AppPreferences::GremlinInfo&	info = gPrefs->GetGremlinInfo ();
			s >> info;
		}
	}

	Platform::UseNewAutoSaveDirectory ();
}


// ---------------------------------------------------------------------------
//		 Gremlins_Dispose
// ---------------------------------------------------------------------------

void Gremlins_Dispose (void)
{
}


// ---------------------------------------------------------------------------
//		 Gremlins_IsOn
// ---------------------------------------------------------------------------

Bool Gremlins_IsOn (void)
{
	return gGremlinsOn;
}


// ---------------------------------------------------------------------------
//		 Gremlins_CanSwitchToApp
// ---------------------------------------------------------------------------

Bool Gremlins_CanSwitchToApp (UInt cardNo, LocalID dbID)
{
	AppPreferences::GremlinInfo&	info = gPrefs->GetGremlinInfo ();

	if (info.fAppList.size () == 0)
	{
		return true;
	}

	AppInfoList::iterator	iter = info.fAppList.begin ();
	
	while (iter != info.fAppList.end ())
	{
		AppInfo	appInfo = *iter++;
		
		if (appInfo.cardNo == cardNo && appInfo.dbID == dbID)
		{
			return true;
		}
	}

	return false;
}


// ---------------------------------------------------------------------------
//		 Gremlins_PostFakeEvent
// ---------------------------------------------------------------------------

Bool Gremlins_PostFakeEvent (void)
{
	return gGremlins.GetFakeEvent() != 0;
}


// ---------------------------------------------------------------------------
//		 Gremlins_PostFakePenEvent
// ---------------------------------------------------------------------------

void Gremlins_PostFakePenEvent (void)
{
	// Only post every once and a while, so as to not overflow
	// the Palm OS's pen queue. (cf. StubEnqueuePenPosition)

	// (Note: taking out the pause.  It lead to indeterminism (the
	//  same Gremlin didn't do exactly the same thing every time),
	//  and doesn't seem necessary anyway, since we only post fake
	//  events when the event queue is empty anyway.)

//	uae_u32	now = Platform::GetMilliseconds() / 20;

//	if (now != gLastTick)
	{
		gGremlins.GetPenMovement ();

//		gLastTick = now;
	}
}


// ---------------------------------------------------------------------------
//		 Gremlins_GremlinNumber
// ---------------------------------------------------------------------------

uae_u32 Gremlins_GremlinNumber (void)
{
	unsigned short	number;
	unsigned long	step;
	unsigned long	until;
	gGremlins.Status (&number, &step, &until);

	return number;
}


// ---------------------------------------------------------------------------
//		 Gremlins_EventCounter
// ---------------------------------------------------------------------------

uae_u32 Gremlins_EventCounter (void)
{
	unsigned short	number;
	unsigned long	step;
	unsigned long	until;
	gGremlins.Status (&number, &step, &until);

	return step;
}


// ---------------------------------------------------------------------------
//		 Gremlins_EventLimit
// ---------------------------------------------------------------------------

uae_u32 Gremlins_EventLimit (void)
{
	unsigned short	number;
	unsigned long	step;
	unsigned long	until;
	gGremlins.Status (&number, &step, &until);

	return until;
}


// ---------------------------------------------------------------------------
//		 Gremlins_ElapsedMilliseconds
// ---------------------------------------------------------------------------

uae_u32 Gremlins_ElapsedMilliseconds (void)
{
	if (Gremlins_IsOn ())
		return Platform::GetMilliseconds () - gGremlinStartTime;

	return gGremlinStopTime - gGremlinStartTime;
}


// ---------------------------------------------------------------------------
//		 Gremlins_New
// ---------------------------------------------------------------------------

void Gremlins_New (void)
{
	AppPreferences::GremlinInfo&	info = gPrefs->GetGremlinInfo ();

	LogStartNew ();

	if (LogGremlins ())
	{
		string	templ = Platform::GetString (kStr_GremlinStarted);
		LogAppendMsg (templ.c_str (), (int) info.fNumber, info.fSteps);
	}

	// If needed, switch to an "approved" application.
	// This code roughly follows that in AppsViewSwitchApp in Launcher.

	if (info.fAppList.size () > 0)
	{
		// Switch to the first on the list.
		
		AppInfo	appInfo = *(info.fAppList.begin ());

		//---------------------------------------------------------------------
		// If this is an executable, call SysUIAppSwitch
		//---------------------------------------------------------------------
		if (IsExecutable (appInfo.type, appInfo.creator, appInfo.dbAttrs))
		{
			Err err = SysUIAppSwitch (appInfo.cardNo, appInfo.dbID,
							sysAppLaunchCmdNormalLaunch, NULL);
			Errors::ThrowIfPalmError (err);
		}

		//---------------------------------------------------------------------
		// else, this must be a launchable data database. Find it's owner app
		//  and launch it with a pointer to the data database name.
		//---------------------------------------------------------------------
		else
		{
			DmSearchStateType	searchState;
			UInt				cardNo;
			LocalID				dbID;
			Err err = DmGetNextDatabaseByTypeCreator (true, &searchState, 
						sysFileTApplication, appInfo.creator, 
						true, &cardNo, &dbID);
			Errors::ThrowIfPalmError (err);

			// Create the param block
			uaecptr	cmdPBP = (uaecptr) MemPtrNew (sizeof (SysAppLaunchCmdOpenDBType));
			Errors::ThrowIfNULL ((void*) cmdPBP);

			// Fill it in
			MemPtrSetOwner ((VoidPtr) cmdPBP, 0);
			put_word (cmdPBP + offsetof (SysAppLaunchCmdOpenDBType, cardNo), appInfo.cardNo);
			put_long (cmdPBP + offsetof (SysAppLaunchCmdOpenDBType, dbID), appInfo.dbID);

			// Switch now
			err = SysUIAppSwitch (cardNo, dbID, sysAppLaunchCmdOpenDB, (Ptr) cmdPBP);
			Errors::ThrowIfPalmError (err);
		}
	}

	gGremlins.Initialize (info.fNumber, info.fSteps);
	gGremlinsOn = true;

	// Make sure we're all on the same page...
	::ResetCalibrationInfo ();

	gPrefs->UseLoggingOptions (AppPreferences::kGremlinLogging);

	gGremlinStartTime = Platform::GetMilliseconds ();

	// Make sure the app's awake.  Normally, we post events on a patch to
	// SysEvGroupWait.  However, if the Palm device is already waiting,
	// then that trap will never get called.  By calling EvtWakeup now,
	// we'll wake up the Palm device from its nap.

	Errors::ThrowIfPalmError (EvtWakeup ());

	Platform::GCW_Open ();
	Platform::UseNewAutoSaveDirectory ();

	if (info.fSaveFrequency != 0)
	{
		regs.spcflags |= SPCFLAG_SAVE_STATE;
	}
}


// ---------------------------------------------------------------------------
//		 Gremlins_Step
// ---------------------------------------------------------------------------

void Gremlins_Step (void)
{
	if (!gGremlinsOn)
	{
		gGremlinsOn = true;
		gGremlins.Step ();

		gPrefs->UseLoggingOptions (AppPreferences::kGremlinLogging);

		// Make sure the app's awake.  Normally, we post events on a patch to
		// SysEvGroupWait.  However, if the Palm device is already waiting,
		// then that trap will never get called.  By calling EvtWakeup now,
		// we'll wake up the Palm device from its nap.

		Errors::ThrowIfPalmError (EvtWakeup ());
	}
}


// ---------------------------------------------------------------------------
//		 Gremlins_Resume
// ---------------------------------------------------------------------------

void Gremlins_Resume (void)
{
	if (!gGremlinsOn)
	{
		gGremlinsOn = true;
		gGremlinStartTime = Platform::GetMilliseconds () - (gGremlinStopTime - gGremlinStartTime);

		// Make sure we're all on the same page...
		::ResetCalibrationInfo ();

		gPrefs->UseLoggingOptions (AppPreferences::kGremlinLogging);

		// Make sure the app's awake.  Normally, we post events on a patch to
		// SysEvGroupWait.  However, if the Palm device is already waiting,
		// then that trap will never get called.  By calling EvtWakeup now,
		// we'll wake up the Palm device from its nap.

		Errors::ThrowIfPalmError (EvtWakeup ());
	}
}


// ---------------------------------------------------------------------------
//		 Gremlins_Stop
// ---------------------------------------------------------------------------

void Gremlins_Stop (void)
{
	if (gGremlinsOn)
	{
		gGremlinsOn = false;
		gGremlinStopTime = Platform::GetMilliseconds ();

		gPrefs->UseLoggingOptions (AppPreferences::kNormalLogging);

		unsigned short	number;
		unsigned long	step;
		unsigned long	until;
		gGremlins.Status (&number, &step, &until);

		if (LogGremlins ())
		{
			string	templ = Platform::GetString (kStr_GremlinEnded);
			LogAppendMsg (templ.c_str (),
				(int) number, step, until, (gGremlinStopTime - gGremlinStartTime));
		}

		LogDump ();
	}
}


// ---------------------------------------------------------------------------
//		 Gremlins_CanNew
// ---------------------------------------------------------------------------

Bool Gremlins_CanNew (void)
{
	return true;
}


// ---------------------------------------------------------------------------
//		 Gremlins_CanStep
// ---------------------------------------------------------------------------

Bool Gremlins_CanStep (void)
{
	return gGremlins.IsInitialized () != 0;
}


// ---------------------------------------------------------------------------
//		 Gremlins_CanResume
// ---------------------------------------------------------------------------

Bool Gremlins_CanResume (void)
{
	return gGremlins.IsInitialized () && !gGremlinsOn;
}


// ---------------------------------------------------------------------------
//		 Gremlins_CanStop
// ---------------------------------------------------------------------------

Bool Gremlins_CanStop (void)
{
	return gGremlinsOn;
}


// -----------------------------
   #pragma mark -
   #pragma mark *** Stub Functions
// -----------------------------



// ---------------------------------------------------------------------------
//		 StubEnqueuePt
// ---------------------------------------------------------------------------
// This is a stub routine that called the application object's method that
// enqueues a point.

void	StubAppEnqueuePt(PointPtr pen)
{
	PointType	pt;
	pt.x = pen->h;
	pt.y = pen->v;
	::EvtEnqueuePenPoint(&pt);
}


// ---------------------------------------------------------------------------
//		 StubEnqueuePt
// ---------------------------------------------------------------------------
// This is a stub routine that called the Application object's method that
// enqueues a key.

void StubAppEnqueueKey (UInt ascii, UInt keycode, UInt modifiers)
{
	switch (ascii)
	{
		// This translates some control characters into system chars
		case 0x01:	// control-A
			ascii = menuChr;
			modifiers = commandKeyMask;
			break;
		case 0x03:	// control-C
			ascii = commandChr;
			modifiers = commandKeyMask;
			break;
		case 0x04:	// control-D
			ascii = confirmChr;
			modifiers = commandKeyMask;
			break;
		case 0x05:	// control-E
			ascii = launchChr;
			modifiers = commandKeyMask;
			break;
		case 0x06:	// control-F
			ascii = keyboardChr;
			modifiers = commandKeyMask;
			break;
		case 0x0D:	// control-D
			ascii = linefeedChr;
			break;
		case 0x0E:	// control-N
			ascii = nextFieldChr;
			modifiers = commandKeyMask;
			break;			
	}


	// If this is one of the hard keys, send it through the Key Manager so that
	//  it can do the right things concering double-taps, etc.
	if ((modifiers & commandKeyMask) &&
			(ascii > hardKeyMin || (ascii == pageUpChr) || (ascii == pageDownChr))) {

		DWord		status = 0;

		switch (ascii) {
			case hard1Chr: 		status = keyBitHard1; break;
			case hard2Chr: 		status = keyBitHard2; break;
			case hard3Chr: 		status = keyBitHard3; break;
			case hard4Chr: 		status = keyBitHard4; break;
			case hardPowerChr: 	status = keyBitPower; break;
			case pageUpChr: 	status = keyBitPageUp; break;
			case pageDownChr: 	status = keyBitPageDown; break;
			}

		// I'm dubious that this is the correct thing to do in the emulator.
		// This is an interrupt routine that we're calling directly.  But
		// what if a "real" interrupt is triggered from some other source
		// while this call is executing (for instance, the user clicks on
		// one of the "keyboard buttons", thus generating a keyboard
		// interrupt)?  It's possible that the two will stomp on each other.
		// The best thing to do is to post events the emulator way.
		// However, I'm hard-pressed to do that.  That would require calling
		// Hardware_ButtonEvent twice (once to signal the keyboard
		// interrupt, and another to clear it), but I don't know how to give
		// the processor time in between to respond to the first call.

		if (status != 0)
			::KeyHandleInterrupt (false, status);
		}

	// For other keys, enqueue them directly.
	else 
		::EvtEnqueueKey (ascii, keycode, modifiers);
}


// ---------------------------------------------------------------------------
//		 StubAppGremlinsOn
// ---------------------------------------------------------------------------
// Stub routine that update the Gremlins menu an the global variable the
// keeps track of the ????

void StubAppGremlinsOn (void)
{
	// Called from Gremlins::Initialize.
}


// ---------------------------------------------------------------------------
//		 StubAppGremlinsOff
// ---------------------------------------------------------------------------
// Stub routine that update the Gremlins menu an the global variable the
// keeps track of the ????

void StubAppGremlinsOff (void)
{
	// Called by Gremlins when counter > until.

	Gremlins_Stop ();
}


// ---------------------------------------------------------------------------
//		 StubViewDrawLine
// ---------------------------------------------------------------------------
// This is a stub routine that called the View object's method that draw
// a line.

void StubViewDrawLine (int xStart, int yStart, int xEnd, int yEnd)
{
	Platform::ViewDrawLine (xStart, yStart, xEnd, yEnd);
}


// ---------------------------------------------------------------------------
//		 StubViewDrawPixel
// ---------------------------------------------------------------------------
// This is a stub routine that called the View object's method that draw
// a pixel.

void StubViewDrawPixel (int xPos, int yPos)
{
	Platform::ViewDrawPixel (xPos, yPos);
}


// ---------------------------------------------------------------------------
//		 operator >> (PPStream&, AppInfo&)
// ---------------------------------------------------------------------------

PPStream& operator >> (PPStream& inStream, AppInfo& outInfo)
{
	LocalID		dbID;
	UInt 		cardNo;
	ULong		modDate;
	Word		category;
	char		name[dmDBNameLength];

	inStream >> outInfo.creator;
	inStream >> outInfo.type;
	inStream >> outInfo.version;

	inStream >> dbID;
	inStream >> cardNo;
	inStream >> modDate;
	inStream >> category;
	inStream >> name;

	return inStream;
}


// ---------------------------------------------------------------------------
//		 operator << (PPStream&, const AppInfo&)
// ---------------------------------------------------------------------------

PPStream& operator << (PPStream& inStream, const AppInfo& inInfo)
{
	LocalID		dbID = 0;
	UInt 		cardNo = 0;
	ULong		modDate = 0;
	Word		category = 0;
	char		name[dmDBNameLength] = {0};

	inStream << inInfo.creator;
	inStream << inInfo.type;
	inStream << inInfo.version;

	inStream << dbID;
	inStream << cardNo;
	inStream << modDate;
	inStream << category;
	inStream << name;

	return inStream;
}


// ---------------------------------------------------------------------------
//		 operator >> (PPStream&, AppPreferences::GremlinInfo&)
// ---------------------------------------------------------------------------

PPStream& operator >> (PPStream& inStream, AppPreferences::GremlinInfo& outInfo)
{
	bool	dummy;

	inStream >> outInfo.fNumber;
	inStream >> outInfo.fSteps;
	inStream >> outInfo.fAppList;
	inStream >> outInfo.fContinuePastWarnings;
	inStream >> dummy;
	inStream >> dummy;

	return inStream;
}


// ---------------------------------------------------------------------------
//		 operator << (PPStream&, const AppPreferences::GremlinInfo&)
// ---------------------------------------------------------------------------

PPStream& operator << (PPStream& inStream, const AppPreferences::GremlinInfo& inInfo)
{
	bool	dummy = false;

	inStream << inInfo.fNumber;
	inStream << inInfo.fSteps;
	inStream << inInfo.fAppList;
	inStream << inInfo.fContinuePastWarnings;
	inStream << dummy;
	inStream << dummy;

	return inStream;
}


// ---------------------------------------------------------------------------
//		 operator >> (PPStream&, AppInfoList&)
// ---------------------------------------------------------------------------

PPStream& operator >> (PPStream& inStream, AppInfoList& outVector)
{
	Int32	numElements;
	inStream >> numElements;

	outVector.clear ();
	outVector.resize (numElements);

	for (Int32 ii = 0; ii < numElements; ++ii)
	{
		inStream >> outVector [ii];
	}

	return inStream;
}


// ---------------------------------------------------------------------------
//		 operator << (PPStream&, const AppInfoList&)
// ---------------------------------------------------------------------------

PPStream& operator << (PPStream& inStream, const AppInfoList& inVector)
{
	Int32	numElements = inVector.size ();

	inStream << numElements;

	for (Int32 ii = 0; ii < numElements; ++ii)
	{
		inStream << inVector [ii];
	}

	return inStream;
}


