/*******************************************************************
 * 							 Pilot Software
 *
 *		 Copyright(c) 1998, Palm Computing Inc., All Rights Reserved
 *
 *-------------------------------------------------------------------
 * FileName:
 *		PPCMgr.c
 *
 * Description:
 *		SerialMgr -> PPCToolbox mapper Routines
 *
 *		These functions are called from SerialMgrMac.c if we're
 *		communicating via the PPCToolbox instead of an actual serial
 *		connection.
 *
 * History:
 *   	6/30/95  kaar - Created
 *
 *******************************************************************/
#define PILOT_PRECOMPILED_HEADERS_OFF

// PalmOS Equates
#include <SysAll.h>
#include <SerialMgr.h>

// PPC Specific Equates
#include "SerLibPPC.h"
#include "PPCMessagingPrv.h"

static void		PrvPPCReconnectPorts (void);
static void		PrvBufferPendingData (void);


// MACRO: checkToYieldCPU
//
// DESCRIPTION:	If at least ticksBetweenYields ticks have expired since
//						the last time we were called, calls a user routine to
//						process Mac events and yield CPU cycles to other tasks
//
// PARAMETERS:		none
//


// We stub out checkToYeildCPU.  It calls a callback function that only
// gets set via a call to SerControl.  However, we've determined that
// SerControl doesn't get called in the PalmDebugger, so the callback
// never gets set, so checkToYieldCPU never has to do anything.

// Actually, we DO need to do something here.  If we're sending a packet
// to the other guy, and we're waiting for a response, we need to give
// up the CPU so that the other guy can process it!  Calling EventAvail
// should do the trick...

// Actually, calling EventAvail doesn't really do the trick.  _Eventually_
// the background application will get called, but only after up to 2000
// calls to EventAvail, which could take .6 seconds. Calling WaitNextEvent
// with a non-zero sleep time has a much better effect (calling WNE with
// a zero sleep time is just as bad as calling EventAvail).

// Also, calling WaitNextEvent to sleep for 1 tick slows down the Debugger's
// startup time dramatically.  Put in a throttle to call it only occassionally.

// Ugh! That fixed the startup time, but now we're back to slow activation
// time to the backgroun app.  Let's try adding a boolean parameter to
// force the call to WaitNextEvent.


#define checkToYieldCPU()							\
	do {											\
		EventRecord aEvent;							\
		(void) WaitNextEvent (0, &aEvent, 1, NULL);	\
	} while (0)


/*
	A small circular buffer class. Bytes can be added and removed, and
	you can find out how many bytes are in there.

	The class uses three pointers.  fBuffer points to the start of the
	memory allocated for the buffer.  fBufferBegin points to the first
	byte in the buffer that can be removed.  fBufferEnd points to the
	first location in the buffer where new bytes can be added.

	Because this is a circular buffer, operations can wrap.  A write
	request may fill up the remaining part of the buffer, wrap, and
	then start filling up the buffer from its beginning.  Similarly
	for read operations.

	A "simple" buffer (one where the stored data is contiguous) looks
	as follows:

	+--------------------------------------------------+
	|              ****************                    |
	+--------------------------------------------------+
	^              ^               ^
	|              |               + fBufferEnd
	|              + fBufferBegin
	+ fBuffer

	A wrapped buffer (one where the stored data is kept in the
	upper and lower portions of the buffer), may look as follows:

	+--------------------------------------------------+
	|*******                                ***********|
	+--------------------------------------------------+
	^       ^                               ^
	|       |                               + fBufferBegin
	|       + fBufferEnd
	+ fBuffer
*/


#if 1
#undef min
#undef max
#include <deque>

class CCircularBuffer
{
	public:
		CCircularBuffer ();
		~CCircularBuffer ();

		void	SetBuffer (void* buffer, long size);

		long	BytesInBuffer (void) const;
		
		void	AddBytes (const void* src, long amt);
		void	RemoveBytes (void* dest, long amt);

	private:
		deque<char>	fBuffer;
};


CCircularBuffer::CCircularBuffer () :
	fBuffer ()
{
}

CCircularBuffer::~CCircularBuffer ()
{
}

void CCircularBuffer::SetBuffer (void*, long)
{
}

long CCircularBuffer::BytesInBuffer (void) const
{
	return fBuffer.size ();
}

void CCircularBuffer::AddBytes (const void* src, long amt)
{
	fBuffer.insert (fBuffer.end (), (char*) src, ((char*) src) + amt);
}

void CCircularBuffer::RemoveBytes (void* dest, long amt)
{
	copy (fBuffer.begin (), fBuffer.begin () + amt, (char*) dest);
	fBuffer.erase (fBuffer.begin (), fBuffer.begin () + amt);
}


#else
class CCircularBuffer
{
	public:
		CCircularBuffer ();
		~CCircularBuffer ();

		void	SetBuffer (void* buffer, long size);

		long	BytesInBuffer (void) const;
		
		void	AddBytes (const void* src, long amt);
		void	RemoveBytes (void* dest, long amt);

	private:
		char*	fBuffer;
		long	fBufferSize;
		char*	fBufferBegin;
		char*	fBufferEnd;
};

CCircularBuffer::CCircularBuffer () :
	fBuffer (NULL),
	fBufferSize (0),
	fBufferBegin (NULL),
	fBufferEnd (NULL)
{
}

CCircularBuffer::~CCircularBuffer ()
{
}

void CCircularBuffer::SetBuffer (void* buffer, long size)
{
	fBuffer = (char*) buffer;
	fBufferSize = size;

	fBufferBegin = fBufferEnd = fBuffer;
}

long CCircularBuffer::BytesInBuffer (void) const
{
	long	amt;

	if (fBufferBegin <= fBufferEnd)
	{
		// Simple case.

		amt = fBufferEnd - fBufferBegin;
	}
	else
	{
		// Wrapped case.

		amt = fBufferSize - (fBufferBegin - fBufferEnd);
	}

	return amt;
}

void CCircularBuffer::AddBytes (const void* src, long amt)
{
	// This assert never lets the buffer get so full that fBufferBegin
	// == fBufferEnd.  When the two are equal, the buffer is *empty*,
	// not completely full.

//	assert (amt < fBufferSize - this->BytesInBuffer ());

	long	amtToEndOfBuffer = (fBuffer + fBufferSize) - fBufferEnd;

	if (amtToEndOfBuffer <= amt)
	{
		// Simple case.

		memcpy (fBufferEnd, src, amt);
		fBufferEnd += amt;
	}
	else
	{
		// Wrapped case.

		memcpy (fBufferEnd, src, amtToEndOfBuffer);

		long	amtRemaining = amt - amtToEndOfBuffer;
		memcpy (fBuffer, ((const char*) src) + amtToEndOfBuffer, amtRemaining);
		fBufferEnd = fBuffer + amtRemaining;
	}
}

void CCircularBuffer::RemoveBytes (void* dest, long amt)
{
	// Make sure we don't bite off more than we got.

//	assert (amt <= this->BytesInBuffer ());

	long	amtToEndOfBuffer = (fBuffer + fBufferSize) - fBufferBegin;

	if (amtToEndOfBuffer <= amt)
	{
		// Simple case.

		memcpy (dest, fBufferBegin, amt);
		fBufferBegin += amt;
	}
	else
	{
		// Wrapped case.

		memcpy (dest, fBufferBegin, amtToEndOfBuffer);

		long	amtRemaining = amt - amtToEndOfBuffer;
		memcpy (((char*) dest) + amtToEndOfBuffer, fBuffer, amtRemaining);
		fBufferBegin = fBuffer + amtRemaining;
	}
}
#endif


CCircularBuffer		gBuffer;
//char				gDefaultBuffer[30*1024];


#pragma mark --- Serial Mgr Intercept Functions ---

/************************************************************
 *
 *  FUNCTION: PPPCOpen
 *
 *  DESCRIPTION: Opens up a serial connection with the given baud rate
 *		and default settings.
 *
 *  PARAMETERS:
 *		connIDP		-> reference to return connection ID
 *		port	 		-> which port to open
 *		baud			-> baud rate to use
 *
 *	CALLED BY:
 *		SerialLink Manager when it opens up a Serial link
 *
 *  RETURNS: 0 if no error; otherwise: serErrBadPort, serErrBadParam
 *
 *  NOTES: the library may be opened only by one client at a time
 *
 *  CREATED: 2/23/95
 *
 *  BY: Vitaly Marty Kruglikov
 *
 *************************************************************/

Err PPPCOpen(void)
{
	// If there's an error, return a non-zero value so that
	// SerOpen will do the serial stuff.

	PPCOpenPorts(kDebuggerPortName, kEmulatorPortName,
						kToEmulatorPortType, kFromEmulatorPortType);

	if (!PPCPortOpened (&gReceivePort))
		return serErrNoMem;

	return errNone;
}


/************************************************************
 *
 *  FUNCTION: PPPCClose
 *
 *  DESCRIPTION: Closes up a serial connection
 *
 *  PARAMETERS:
 *		connID		 	-> connection ID
 *
 *	CALLED BY:
 *		Whoever wants to close a serial connection
 *
 *  RETURNS: 0 if no error
 *
 *  CREATED: 2/23/95
 *
 *  BY: Vitaly Marty Kruglikov
 *
 *************************************************************/

Err PPPCClose(void)
{
	PPCClosePorts ();

	return errNone;
}


/************************************************************
 *
 *  FUNCTION: PPCClearErr
 *
 *  DESCRIPTION: Clears current error out of serial port
 *
 *  PARAMETERS:
 *		conID		 	-> connection ID
 *
 *	CALLED BY:
 *
 *  RETURNS: 0 if no error
 *
 *  CREATED: 2/23/95
 *
 *  BY: Ron Marianetti
 *
 *************************************************************/

Err PPCClearErr(void)
{
	return errNone;
}


/************************************************************
 *
 *  FUNCTION: PPCSendWait
 *
 *  DESCRIPTION: Waits for all characters in the send buffer to
 *		be sent over the serial port or a timeout, whichever comes
 *		sooner.
 *
 *  PARAMETERS:
 *		refNum	 		-> lib refnum
 *		timeout			-> IGNORED: the port's CTS timeout value is used instead!!!
 *
 *	CALLED BY:
 *
 *  RETURNS: 0 if no error; serErrBadConnID, serErrTimeOut
 *
 *  CREATED: 2/23/95
 *
 *  BY: Vitaly Marty Kruglikov
 *
 *			vmk	1/29/96	Converted to use CTS timeout like the native version
 *
 *************************************************************/

Err PPCSendWait(Long timeout)
{
#pragma unused (timeout)

	ULong	startTicks;
	Err		err = errNone;

	// Check for and deal with unconnected ports.
	PrvPPCReconnectPorts ();

	startTicks = TimGetTicks();					// init starting ticks

	// get output driver status

	while (PPCPortWriting (&gSendPort) &&
		((TimGetTicks() - startTicks) < serDefaultCTSTimeout))
	{
		checkToYieldCPU();			// yield processor
	}

	if (PPCPortWriting (&gSendPort))
		err = serErrTimeOut;

	if (!PPCPortConnected (&gSendPort))
		err = serErrNotOpen;

	return err;
}


/************************************************************
 *
 *  FUNCTION: PPCReceiveWait
 *
 *  DESCRIPTION: Waits for 'size' bytes of data to arrive at the
 *		serial port, or for a timeoout
 *
 *  PARAMETERS:
 *		conID		 	-> connection ID
 *		bytes			-> number of bytes to wait for
 *		timeout			-> timeout in ticks, 0 for none, -1 forever
 *
 *	CALLED BY:
 *
 *
 *  RETURNS: 0 if no error
 *
 *  CREATED: 2/23/95
 *
 *  BY: Vitaly Marty Kruglikov
 *
 *************************************************************/

Err PPCReceiveWait(ULong bytes, Long timeout)
{
	ULong	startTicks;		// starting tick count

	if (!bytes)
		return errNone;

	// Check for and deal with unconnected ports.

	PrvPPCReconnectPorts ();

	// If there's no timeout value && there's no data, blow out.
	// We want to do this _before_ calling checkToYieldCPU, as
	// that call can take some time, which we don't want to expend
	// on a zero-timeout call.

	if (!timeout && !gReceivePort.fDataWaiting)
	{
		return serErrTimeOut;
	}

	// Wait for some data to arrive keeping track of the timeout value

	startTicks = TimGetTicks();		// get initial tick count

	PrvBufferPendingData ();

	while (gBuffer.BytesInBuffer () < bytes)
	{
		if (timeout != -1 && (TimGetTicks() - startTicks) >= timeout)
		{
			return serErrTimeOut;			// timed out
		}

		checkToYieldCPU();				// yield processor

		PrvBufferPendingData ();
	}

	// Return that we have the requested data.

	return errNone;
}


/************************************************************
 *
 *  FUNCTION: PPCReceiveFlush
 *
 *  DESCRIPTION: Flushes all characters in the receive buffer and
 *		all characters coming into the serial port until no
 *		more arrive within the timeout period.
 *
 *  PARAMETERS:
 *		conID		 	-> connection ID
 *		timeout			-> timeout in ticks, 0 for none, -1 forever
 *
 *	CALLED BY:
 *
 *
 *  RETURNS: 0 if no error
 *
 *  CREATED: 2/23/95
 *
 *  BY: Vitaly Marty Kruglikov
 *
 *************************************************************/

void PPCReceiveFlush(Long timeout)
{
#pragma unused (timeout)
}


/************************************************************
 *
 * FUNCTION:	PPCSetReceiveBuffer
 *
 * DESCRIPTION:	Specify a new input buffer.
 *
 *						To restore the original buffer, pass bufSize = 0.
 *
 * PARAMETERS:	refNum		-- serial library reference number
 *					bufP			-- pointer to the new buffer(must be locked)
 *					bufSize		-- size, in bytes, of the new input buffer.  Set
 *									   this value to 0 to restore the default buffer.
 * CALLED BY:
 *
 * RETURNS: 0 if no error
 *
 * CREATED: 8/21/95
 *
 * BY: Vitaly Marty Kruglikov
 *
 *************************************************************/

Err PPCSetReceiveBuffer(VoidPtr bufP, UInt bufSize)
{
#pragma unused (bufP, bufSize)

/*
	if (!bufSize)
	{
		bufP = gDefaultBuffer;
		bufSize = sizeof (gDefaultBuffer);
	}

	gBuffer.SetBuffer (bufP, bufSize);
*/

	return errNone;
}


/************************************************************
 *
 *  FUNCTION: PPCSend (v2.0)
 *
 *  DESCRIPTION: Sends 1 or more bytes of data over the serial port
 *
 *  PARAMETERS:
 *		refNum		->		serial lib refNum
 *		bufP			->		pointer to data to send
 *		count			->		# of bytes to send
 *		errP			<->	for returning error code
 *
 *	CALLED BY:
 *		Whoever wants to send data via the serial port
 *
 *  RETURNS: number of bytes transferred
 *
 *  *errP =		0						-- no error
 *					serErrTimeOut		-- handshake timeout
 *
 *  CREATED: 2/23/95
 *
 *  BY: Vitaly Marty Kruglikov
 *
 *			vmk	7/31/96	Extended the API to return the number of bytes sent
 *
 *************************************************************/

ULong PPCSend(VoidPtr bufP, ULong count, Err* errP)
{
	ULong		startTicks;

	// Check for and deal with unconnected ports.
	PrvPPCReconnectPorts ();

	*errP = errNone;

	if (!PPCPortConnected (&gSendPort))
	{
		*errP = serErrNotOpen;
		return 0;
	}

	PPCSendData (bufP, count);

	startTicks = TimGetTicks();
	while (PPCPortWriting (&gSendPort))
	{
		checkToYieldCPU();		// yield processor

		if ((TimGetTicks() - startTicks) >= serDefaultCTSTimeout)
		{
			*errP = serErrTimeOut;
			return 0;
		}
	}

	if (!PPCPortConnected (&gSendPort))
	{
		*errP = serErrNotOpen;
		return 0;
	}

	return count;
}


/************************************************************
 *
 *  FUNCTION: PPCReceive (v2.0)
 *
 *  DESCRIPTION: Receives 'count' bytes worth of data or returns with
 *		error if a line error or timeout is encountered
 *
 *  PARAMETERS:
 *		refNum		->		serial lib refNum
 *		rcvBufP		<->	buffer for receiving data
 *		count			->		number of bytes to receive
 *		timeout		->		inter-byte timeout in ticks, 0 for none, -1 forever
 *
 *	CALLED BY:
 *
 *
 *  RETURNS: number of bytes received
 *
 *		*errP =	0					-- no error
 *					serErrLineErr	-- RS232 line error
 *					serErrTimeOut	-- inter-byte timeout
 *
 *  CREATED: 2/23/95
 *
 *  BY: Vitaly Marty Kruglikov
 *
 *		vmk	8/1/96		Extended the API to return the number of bytes received
 *
 *************************************************************/

ULong PPCReceive(VoidPtr bufP, ULong count, Long timeout, Err* errP)
{
	ULong	received = 0;
	*errP = errNone;

	// Check for and deal with unconnected ports.

	PrvPPCReconnectPorts ();

	// Receive any pending data.

	PrvBufferPendingData ();

	// If it's not enough, wait for more.

	if (gBuffer.BytesInBuffer () < count)
	{
		*errP = PPCReceiveWait (count, timeout);

		if (!*errP)
		{
			PrvBufferPendingData ();
		}
	}

	// If we have all the needed bytes, return the requested amount.

	if (*errP == errNone)
	{
		gBuffer.RemoveBytes (bufP, count);
	}

	return count;
}


// ---------------------------------------------------------------------------
//		 PrvPPCOpenPorts
// ---------------------------------------------------------------------------
// Check for and deal with unconnected ports.

void PrvPPCReconnectPorts (void)
{
	if (PPCPortOpened (&gReceivePort) &&
		!PPCPortConnecting (&gReceivePort) &&
		!PPCPortConnected (&gReceivePort))
	{
		PPCClosePorts ();
		PPCOpenPorts (kDebuggerPortName, kEmulatorPortName,
						kToEmulatorPortType, kFromEmulatorPortType);
	}
}


// ---------------------------------------------------------------------------
//		 PrvBufferPendingData
// ---------------------------------------------------------------------------

void PrvBufferPendingData (void)
{
	if (gReceivePort.fDataWaiting)
	{
		gReceivePort.fDataWaiting = 0;

		void*	inBuffer = gReceivePort.fPB.readParam.bufferPtr;
		long	inAmt = gReceivePort.fPB.readParam.actualLength;

		gBuffer.AddBytes (inBuffer, inAmt);
		PPCListenForData (&gReceivePort);
	}
}
