/*********************************************************************
*
* Copyright (c) 1998
* 3Com/Palm Computing Division.  All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions
* are met:
*
* 1. Redistributions of source code must retain the above copyright
*	 notice, this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above
*	 copyright notice, this list of conditions and the following
*	 disclaimer in the documentation and/or other materials provided
*	 with the distribution.
*
* 3. All advertising materials mentioning features or use of this
*	 software must display the following acknowledgement:
*
*		This product includes software developed by 3Com and its
*		contributors.
*
* 4. Neither 3Com nor the names of its contributors may be used to
*	 endorse or promote products derived from this software
*	 without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE 3COM AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
* TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 3COM OR
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
* USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
* OF SUCH DAMAGE.
*
*-------------------------------------------------------------------
* FileName:
*	SerLibSocketsMac.cpp
* 
* Description:
*
*	This source file implements the routines used by SerTransport.cpp
*	when the serial transport mode is set as tptSockets (as defined in
*	SerTransport.h). 
*
*	The SerTransport routines are used in place of the SerXXX routines
*	when running in a host application like PalmDebugger or PalmOSEmulator.
*	These routines allow us to generically replace the serial comm
*	with an alternative comm.  
*
*	The SerialLinkMgr.c source file contains an #ifdef near the top
*	that remaps all the SerXXX routines to TptXXX routines. Each of
*	the TptXXX routines contains run-time checks that dispatch to
*	the appropriate low level comm handler routine depending on the
*	current transport setting (stored in a global and setup by TptSetType()
*
* Notes:
*	An feeble attempt has been made to make this source file buildable on any
*	  platform that has Berkeley sockets API calls but for a number of
*	  reasons it's only Win32 for now. 
*
* History:
*	21-Nov-1998  - Created from SocketMessaging.cpp
*
****************************************************************/

// Various Transports
#include "SerLibSockets.h"


/***********************************************************************
*
* FUNCTION: 	SerSockStart
*
* DESCRIPTION:	Starts up the sockets mechanism for communicating with
*				the debugger.  It first attempts to connect to the
*				debugger, assuming that it's been started up and issued
*				a listen().  If that fails, it creates a socket to
*				listen on itself.
*
* PARAMETERS:	
*				connecting	  IN	if true, connect to other host. 
*									if false, creating listening socket
*
* RETURNED: 	0 if no err.
*
* CALLED BY:	TptOpen() in SerTransport.cpp
*
* NOTES:
*
***********************************************************************/
long SerSockStart (bool connecting)
{
	connecting = connecting;
	return 0;
}


/***********************************************************************
*
* FUNCTION: 	SerSockStop
*
* DESCRIPTION:	Stops the sockets session.	Closes all sockets, sets
*				our state to unconnected, and closes down the sockets
*				library (on Windows).
*
* PARAMETERS:	None.
*
* RETURNED:		0 if no err.
*
* CALLED BY:	TptClose() in SerTransport.cpp
*
* NOTES:
*
***********************************************************************/

long SerSockStop (void)
{
	return 0;
}


/***********************************************************************
*
* FUNCTION: SerSockConnectPending
*
* DESCRIPTION:	Returns whether or not there is a pending connect.	This
*				function can be called at any time.  It will only return
*				true if there is a non-blocking listen issued with a
*				pending connect that we haven't yet accepted.
*
* PARAMETERS:	None.
*
* RETURNED: 	True if there is a pending connect.  If there is, the
*				caller should next call SerSockAcceptConnection.
*
* CALLED BY:	PalmOSEmulator in it's idle loop
*
* NOTES:
*
***********************************************************************/

bool SerSockConnectPending (void)
{
	return false;
}


/***********************************************************************
*
* FUNCTION: 	SerSockAcceptConnection
*
* DESCRIPTION:	Accepts a pending connection.  Assumes that
*				ConnectPending has already been called and returned
*				true.  After this, the socket connection should be open
*				for business.
*
* PARAMETERS:	None.
*
* RETURNED: 	0 if no error.
**
* CALLED BY:	PalmOSEmulator in it's idle loop
*
* NOTES:
*
***********************************************************************/

long SerSockAcceptConnection (void)
{
	return -1;
}


/***********************************************************************
*
* FUNCTION: SerSockIsConnected
*
* DESCRIPTION:	Returns whether or not we have a live connection to an
*				external debugger via a socket.  This function can be
*				called at any time.
*
* PARAMETERS:	None.
*
* RETURNED: Nothing.
*
*
* CALLED BY:	PalmOSEmulator in it's idle loop
*
* NOTES:
*
***********************************************************************/
bool SerSockIsConnected (void)
{
	return false;
}




/***********************************************************************
*
* FUNCTION: SerSockSend
*
* DESCRIPTION:	Writes data out to the buffer.
*
* PARAMETERS:	buffer - pointer to the data to be written.
*
*				amountToWrite - number of bytes in the buffer.
*
*				errP - error code returned here
*
* RETURNED:		# of bytes actually written
*
*				On any error, the sub-system will be closed down.  All
*				sockets will be closed and the state will be set to
*				unconnected.
*
* CALLED BY:	TptSend() in SerTransport.cpp
*
* NOTES:
*
***********************************************************************/
long SerSockSend (const void* buffer, long amountToWrite, short* errP)
{
	buffer = buffer;
	amountToWrite = amountToWrite;
	errP = errP;
	return 0;
}


/***********************************************************************
*
* FUNCTION:		SerSockReceivePartial
*
* DESCRIPTION:	Receive up to count bytes from the socket.	It is assumed that 
*				HasUnreadData has already been called and returned true.
*
* PARAMETERS:	buffer	- pointer to the buffer to put the data.
*
*				count	-	 size of the buffer.
*				timeout	- timeout, -1 means wait forever
*				errP	- error returned here
*
* RETURNED: The actual number of bytes read, or -1 on error
*
*
* CALLED BY:	Apps that wish to read as much data as possible
*
* NOTES:
*
***********************************************************************/

long SerSockReceivePartial (void* buffer, unsigned long count, long timeout,
								short* errP)
{
	buffer = buffer;
	count = count;
	timeout = timeout;
	errP = errP;
	return 0;
}


/***********************************************************************
*
* FUNCTION: SerSockReceive
*
* DESCRIPTION:	Don't return until count bytes of data have been read
*				from the socket, or until an error occurs. 
*
* PARAMETERS:	buffer	  IN  pointer to the buffer to put the data.
*				count	  IN  number of bytes to read
*				timeout   IN  timeout in ticks
*				*errP	  OUT error, if any returned here
*
* RETURNED: The actual number of bytes read.
*
*
* CALLED BY:	TptReceive() in SerTransport.cpp
*
* NOTES:
*
***********************************************************************/
long	SerSockReceive	(void* buffer, unsigned long count, long timeout, 
							 short* errP)
{
	buffer = buffer;
	count = count;
	timeout = timeout;
	errP = errP;
	return 0;
}



/***********************************************************************
*
* FUNCTION: SerSockReceiveWait
*
* DESCRIPTION:	Don't return until at least 1 byte is available
*					on the socket, or until error occurrs
*
* PARAMETERS:	
*				bytes	  IN  ignored
*				timeout   IN  timeout in ticks, -1 means wait forever
*
* RETURNED: 0 if no error
*
*
* CALLED BY:	TptReceiveWait() in SerTransport.cpp
*
* NOTES:
*
***********************************************************************/
long	SerSockReceiveWait	(unsigned long bytes, long timeout)
{
	bytes = bytes;
	timeout = timeout;
	return -1;
}


/***********************************************************************
*
* FUNCTION: SerSockReceiveFlush
*
* DESCRIPTION:	Flush all data available in socket 
*
* PARAMETERS:	
*				bytes	  IN  ignored
*				timeout   IN  not used
*
* RETURNED: 0 if no error
*
*
* CALLED BY:	TptReceiveFlush() in SerTransport.cpp
*
* NOTES:
*
***********************************************************************/
long	SerSockReceiveFlush ( long timeout)
{
	timeout = timeout;
	return 0;
}



