/*******************************************************************
* 							 Touchdown Software
*
*		 Copyright(c) 1995, Palm Computing Inc., All Rights Reserved
*
*-------------------------------------------------------------------
* FileName:
*		SerMgr.c
*
* Description:
*		Windows95 version of the TouchDown serial-line-level communication routines.
*
* History:
*   	11/5/97  KK - Created by Karen Kurreck
*
*******************************************************************/


// Include windows headers
#include <windows.h>
#include <crtdbg.h>			// _CrtCheckMemory()
#include <stdio.h>
#include <assert.h>

// UI Equates for transport identifiers
#include "UIGlue.h"

// PalmOS equates
#include "SerLibWinPrv.h"
#include <ErrorMgr.h>


//
// Serial library globals
//
static SerGlobalsPtr	PrvGlobalsP;

// saved reference num for error checking
static UInt				PrvSavedRefNum = 0; 



BOOL PrvSerGetQueueStats(DWord* pcbInQue, DWord* pcbOutQue);

#if 0
#define translateLineErrors(resultP, serLineErrs)					\
	/* Translates Windows GetCommError errors  into our own line */	\
	/* errors bitfield format appropriate for the */				\
	/* Serial Manager globals lineErrors field. */					\
	{																			\
	/* initialize result */											\
	*(resultP) = 0;													\
																	\
	/* software overrun ? */										\
	if ( (serLineErrs) & (CE_TXFULL | CE_RXOVER) )					\
	*(resultP) |= serLineErrorSWOverrun;							\
																	\
	/* hardware overrun ? */										\
	if ( (serLineErrs) & (CE_OVERRUN) )								\
	*(resultP) |= serLineErrorHWOverrun;							\
																	\
	/* break signal asserted ? */									\
	if ( (serLineErrs) & CE_BREAK )									\
	*(resultP) |= serLineErrorBreak;								\
																	\
	/* parity error ? */											\
	if ( (serLineErrs) & CE_RXPARITY )								\
	*(resultP) |= serLineErrorParity;								\
																	\
	/* framing error ? */											\
	if ( (serLineErrs) & CE_FRAME )									\
	*(resultP) |= serLineErrorFraming;								\
																	\
	/* line hand-shake error ? */									\
	if ( (serLineErrs) & (CE_CTSTO | CE_DSRTO | CE_RLSDTO) )		\
	*(resultP) |= serLineErrorHShake;								\
	}
#endif

#define serDefaultRcvBufferSize		(30*1024)
#define serDefaultXmitBufferSize	(2*1024)


///////////////////////////////////////////////////////////////////////////
//
//	Debugger serial I/O routines
//
///////////////////////////////////////////////////////////////////////////


/************************************************************
*
*  FUNCTION: SerOpen
*
*  DESCRIPTION: Opens up a serial connection with the given baud rate
*		and default settings.
*
*  PARAMETERS: 
*		refNum  		-- serial library ref number
*		port	 		-- which port to open (COM1 - COM4)
*		baud			-- baud rate to use
*
*	CALLED BY:
*		SerialLink Manager when it opens up a Serial link or
*		called directly by Debugger
*
*  RETURNS: 0 if no error; 
*      otherwise: serErrBadPort, serErrBadParam, serErrAlreadyOpen
*
*  NOTES: the library may be opened only by one client at a time
*
*
*************************************************************/
Err SerOpen(UInt refNum, UInt port, ULong baud)
{

  SerGlobalsPtr	gP = PrvGlobalsP;
  Err				err = serErrBadPort;
  SerSettingsType	serSettings;
  char strport[32];

  
  // If no globals, initialize them
  if (!gP)
	{
	  gP = (SerGlobalsPtr)new char [sizeof (SerGlobalsType)];
	  if (gP) memset (gP, 0, sizeof (SerGlobalsType));
	  PrvGlobalsP = gP;
	}
  
  ErrFatalDisplayIf( !gP, "nil serial manager globals" );
  ErrFatalDisplayIf( (gP->openCount > 1), "invalid serial manager open count" );
  ErrFatalDisplayIf( gP->openCount, "serial manager is in use" );
  if ( !gP )	return ( serErrNoMem );
  if ( gP->openCount ) return( serErrAlreadyOpen );
  
  // open the comm port if port >=1
  // Or open a pipe if port = 0
  if (port) 
	{
	  sprintf( strport, "COM%d", (int)port );
	  gP->commHdl = CreateFile(strport,
		  GENERIC_READ | GENERIC_WRITE,
		  0,             // exclusive access
		  NULL,          // no security attrs
		  OPEN_EXISTING,
		  FILE_ATTRIBUTE_NORMAL,  
		  (HANDLE)NULL);
	  
	} 
  else 
	{
	  // Open or create a pipe
	  gP->commHdl = CreateFile("\\\\.\\pipe\\Emulator_Debugger",
		  GENERIC_READ | GENERIC_WRITE,
		  0,             // exclusive access
		  NULL,          // no security attrs
		  OPEN_EXISTING,
		  FILE_ATTRIBUTE_NORMAL,  
		  (HANDLE)NULL);
	  if ( gP->commHdl == (HANDLE) INVALID_HANDLE_VALUE ) {
		  // Pipe not existing, so create it
		  gP->commHdl = CreateNamedPipe( 
			  "\\\\.\\pipe\\Emulator_Debugger",             // pipe name 
			  PIPE_ACCESS_DUPLEX, //|      // read/write access 
			  PIPE_TYPE_BYTE |       // message-type pipe 
			  PIPE_READMODE_BYTE |   // message read mode 
			  PIPE_WAIT,                // blocking mode 
			  PIPE_UNLIMITED_INSTANCES, // unlimited instances 
			  serDefaultXmitBufferSize,                  // output buffer size 
			  serDefaultRcvBufferSize,                  // input buffer size 
			  1000,             // client time-out in ms
			  NULL);                    // no security attributes
		}
	  
	}


  if ( gP->commHdl == (HANDLE) INVALID_HANDLE_VALUE )
	{
	  DWord		commErr;
	  commErr = GetLastError();
	  return( serErrBadPort );// error opening port
	}
		
  gP->inQSize = serDefaultRcvBufferSize;
  gP->outQSize = serDefaultXmitBufferSize;
  gP->openCount = 1;
  PrvSavedRefNum = refNum;		// save ref num for error checking
  serSettings.baudRate = baud;
  serSettings.flags = serDefaultSettings;
  serSettings.ctsTimeout = serDefaultCTSTimeout;
  
  // enable hardware handshake at higher baud rates
  if (baud >= 19200L)
	  serSettings.flags |= serSettingsFlagCTSAutoM;
  gP->settings = serSettings;			// initialize settings to prevent recursion (see SerSetSettings)
  err = SerSetSettings( refNum, &serSettings );
  if ( err )
	  SerClose( refNum );
  
  return( err );
}



/************************************************************
*
*  FUNCTION: SerClose
*
*  DESCRIPTION: Closes up a serial connection
*
*  PARAMETERS: 
*		refNum		-- serial library ref number
*
*	CALLED BY:
*		Whoever wants to close a serial connection
*
*  RETURNS: 0 if no error; otherwise serErrBadPort
*
*************************************************************/
//DllExport 
Err SerClose(UInt refNum)
{
  SerGlobalsPtr	gP = PrvGlobalsP;
  Err				err = 0;
  BOOL			fSuccess;
  
  ErrFatalDisplayIf( (gP->openCount == 0), "serial manager was not open" );
  ErrFatalDisplayIf( (gP->openCount > 1), "invalid serial manager open count" );
  ErrFatalDisplayIf( (refNum != PrvSavedRefNum), "invalid ref num" );
  
  gP->openCount = 0;
  
  if (gP->port) {
	  // We were speaking through serial port 1..4
	  // drop DTR
	  fSuccess = EscapeCommFunction(gP->commHdl, CLRDTR);
	  
	  // Purge any outstanding reads/writes and close device handle
	  if (fSuccess)
		  fSuccess = PurgeComm(gP->commHdl, PURGE_TXABORT | PURGE_RXABORT |
		  PURGE_TXCLEAR | PURGE_RXCLEAR);
	  if (fSuccess)
		  fSuccess = CloseHandle(gP->commHdl);
	  if (!fSuccess)
		  err = serErrBadPort;
	  else gP->commHdl = NULL;
	  } else {
	  // We were using a pipe to communicate (port = 0)
	  fSuccess = DisconnectNamedPipe( gP->commHdl);
	  fSuccess |= CloseHandle(gP->commHdl);
	  if (!fSuccess)
		  err = serErrBadPort;
	  
		  }
	  return( err );
}



/************************************************************
 *
 *  FUNCTION: SerSetMapPort
 *
 *  DESCRIPTION: Setup the mapping of the pilot serial port
 *
 *  PARAMETERS: 
 *		pilotPort	-> pilot port number
 *		macPort	 	-> mac port number
 *
 *	CALLED BY:
 *		Any mac application that wants to redirect the emulated pilot
 *		serial port to a specific mac serial port.
 *
 *  RETURNS: the previous mapping
 *
 *  CREATED: 12/9/96 
 *
 *  BY: Gregory Toto
 *
 *************************************************************/
UInt SerSetMapPort( UInt pilotPort, UInt macPort )
{
  return 0;	
}	// SerSetMapPort



/************************************************************
 *
 * FUNCTION:	SerSetReceiveBuffer
 *
 * DESCRIPTION:	Specify a new input buffer.
 *
 *						To restore the original buffer, pass bufSize = 0.
 *
 * PARAMETERS:	refNum		-- serial library reference number
 *					bufP			-- pointer to the new buffer(must be locked)
 *					bufSize		-- size, in bytes, of the new input buffer.  Set
 *									   this value to 0 to restore the default buffer.
 * CALLED BY:
 *		
 * RETURNS: 0 if no error
 *
 * CREATED: 8/21/95 
 *
 * BY: Vitaly Marty Kruglikov
 *
 *************************************************************/
Err SerSetReceiveBuffer(UInt refNum, VoidPtr bufP, UInt bufSize)
{
	return( 0 );
}


/************************************************************
*
*  FUNCTION: SerSetSettings
*
*  DESCRIPTION: Changes settings of serial port
*
*  PARAMETERS: 
*		refNum		-- serial library ref number
*		settingsP	-- new settings
*
*	CALLED BY:	anyone
*		
*
*  RETURNS: 0 if no error; otherwise: 
*		serErrBadPort, serErrBadParam, serErrIOError
*
*
*************************************************************/
Err SerSetSettings(UInt refNum,SerSettingsPtr settingsP)
{				   
  SerGlobalsPtr	gP = PrvGlobalsP;
  DCB				dcb;				// device control block
  COMMTIMEOUTS	cto;
  DWord			commErr = 0;
  Err				retval = serErrLineErr;
  Boolean			fSuccess;
  

  ErrFatalDisplayIf( (gP->openCount == 0), "serial manager is not open" );
  ErrFatalDisplayIf( (refNum != PrvSavedRefNum), "invalid ref num" );
  
  // KK: Note: MAC/Pilot code...
  // Win32 seems to be able to do this without opening and closing port...
  
  // ***There appears to be an error when switching settings from
  // ***hardware handshaking to no hardware handshaking.  When this
  // ***occurs, the transmitter does not want to transmit any more.
  // ***As a workaround, we are going to close and re-open the port
  // ***when we detect this.  This should be transparent to our caller.
  //
  //if (  (gP->settings.flags & serSettingsFlagCTSAutoM) &&
  //		!(settingsP->flags & serSettingsFlagCTSAutoM) )
  //	{
  //	Err	err;
  //	int	savedPort;
  //	ULong	savedBaud;
  
  //	savedPort = gP->port;						// save original settings for re-opening
  //	savedBaud = gP->baudRate;
  
  //	SerClose( refNum );							// close and re-open the port
  //	err = SerOpen( refNum, savedPort, savedBaud );
  //	if ( err )
  //		return( err );
  //	}
  
  if (gP->commHdl)
	  {
	  if (SetupComm(gP->commHdl, gP->inQSize, gP->outQSize))
		  {
		  switch ( settingsP->baudRate )
			  {
			  case 300L:
			  case 600L:
			  case 1200L:
			  case 2400L:
			  case 4800L:
			  case 9600L:
			  case 14400L:
			  case 19200L:
			  case 38400L:
			  case 57600L:
			  case 115200L:
			  case 230400L:
				  break;
			  default:
				  {return( serErrBadParam );}
			  }	
		  
		  // set timeout
		  if (settingsP->baudRate <= CBR_2400)
			  {
			  cto.ReadIntervalTimeout         = 2*1000;  
			  cto.ReadTotalTimeoutMultiplier  = 15;  
			  cto.ReadTotalTimeoutConstant    = 500; 
			  cto.WriteTotalTimeoutMultiplier = 15;  
			  cto.WriteTotalTimeoutConstant   = 500; 
			  }
		  else
			  {
			  cto.ReadIntervalTimeout         = 3*1000;
			  cto.ReadTotalTimeoutMultiplier  = 25;
			  cto.ReadTotalTimeoutConstant    = 2000;
			  cto.WriteTotalTimeoutMultiplier = 25;
			  cto.WriteTotalTimeoutConstant   = 2000;
			  }
		  
		  if (!SetCommTimeouts(gP->commHdl, &cto))
			  {
			  commErr = GetLastError();
			  }
		  else
			  {
			  dcb.DCBlength = sizeof(DCB);
			  // note: this will stuff dcb with current values...
			  fSuccess = GetCommState(gP->commHdl, &dcb);
			  if ( !fSuccess )
				  commErr = GetLastError();
			  
			  // set new baud rate
			  dcb.BaudRate = (UINT)settingsP->baudRate;
			  // Determine parity
			  if ( settingsP->flags & serSettingsFlagParityOnM )
				  {
				  if (settingsP->flags & serSettingsFlagParityEvenM)
					  dcb.Parity = EVENPARITY;		// even parity
				  else
					  dcb.Parity = ODDPARITY;			// odd parity
				  }
			  else
				  {
				  dcb.Parity = NOPARITY;	// no parity
				  }
			  dcb.fParity     = FALSE; 
			  
			  // Determine stop bits
			  if (settingsP->flags & serSettingsFlagStopBits2 )
				  dcb.StopBits = TWOSTOPBITS; // 2 stop bits
			  else
				  dcb.StopBits = ONESTOPBIT;	// 1 stop bit
			  
			  // Determine data bits
			  switch ( settingsP->flags & serSettingsFlagBitsPerCharM )
				  {
				  case serSettingsFlagBitsPerChar5:
					  dcb.ByteSize = 5;
					  break;
				  case serSettingsFlagBitsPerChar6:
					  dcb.ByteSize = 6;
					  break;
				  case serSettingsFlagBitsPerChar7:
					  dcb.ByteSize = 7;
					  break;
				  case serSettingsFlagBitsPerChar8:
					  dcb.ByteSize = 8;
					  break;
				  default:
					  {return( serErrBadParam );}
					  //break;
				  }
			  // Determine flow control
			  dcb.fBinary     = TRUE;
			  if ( settingsP->flags & serSettingsFlagXonXoffM )
				  {
				  dcb.fOutX       = 1;
				  dcb.fInX        = 1;
				  }
			  else
				  {
				  dcb.fOutX       = 0;
				  dcb.fInX        = 0;
				  }
			  dcb.fOutxDsrFlow = FALSE;	
			  dcb.fDtrControl  = DTR_CONTROL_ENABLE;
			  
			  // HW Handshake
			  if ( settingsP->flags & (serSettingsFlagCTSAutoM) )
				  {
				  dcb.fOutxCtsFlow = TRUE;              
				  dcb.fRtsControl  = RTS_CONTROL_HANDSHAKE;
				  }
			  else
				  {
				  dcb.fOutxCtsFlow = FALSE;
				  dcb.fRtsControl  = RTS_CONTROL_ENABLE; 
				  }
			  fSuccess = SetCommState(gP->commHdl, &dcb);
			  if ( !fSuccess )
				  commErr = GetLastError();
			  else  retval  = 0;
			  } // no SetCommTimeouts error
  } // no SetupComm error
  else
	  {
	  commErr = GetLastError();
	  }
} // if commHdl

// Save the settings
gP->settings = *settingsP;	

return retval;
}

/************************************************************
*
*  FUNCTION: SerGetSettings
*
*  DESCRIPTION: Retreives settings of serial port
*
*  PARAMETERS: 
*		refNum		 	-> serial library reference number
*		settingsP 		-> settings
*
*	CALLED BY:	anyone
*		
*************************************************************/
Err SerGetSettings (UInt refNum, SerSettingsPtr settingsP)
{
  SerGlobalsPtr	gP = PrvGlobalsP;
  
  ErrFatalDisplayIf( !gP->openCount, "serial manager is not open" );
  ErrFatalDisplayIf( (refNum != PrvSavedRefNum), "invalid ref num" );
  
  *settingsP = gP->settings;
  
  return 0;
}



/************************************************************
*
*  FUNCTION: SerClearErr
*
*  DESCRIPTION: Clears current error out of serial port
*
*  PARAMETERS:	refNum		 -> library reference number
*
*	CALLED BY:	anyone
*		
*  RETURNS:	0 on success
*
*
*************************************************************/
//DllExport
Err SerClearErr (UInt refNum)
{
  
  SerGlobalsPtr	gP = PrvGlobalsP;
  DWord   dwErrorMask = (DWord)0;
  
  ErrFatalDisplayIf( !gP->openCount, "must open serial manager first" );
  ErrFatalDisplayIf( (refNum != PrvSavedRefNum), "invalid ref num" );
  
  // This call clears comm errors on the device
  ClearCommError( gP->commHdl, &dwErrorMask, NULL /* lpStat */);
  
  return 0;
}



/************************************************************
*
*  FUNCTION: SerSend
*
*  DESCRIPTION: Sends 1 or more bytes of data over the serial port
*      Note - counts on Win32 WriteFile to check timeout set in 
*      SetCommTimeouts (called in SerSetSettings)
*
*  PARAMETERS: 
*		refNum		-> serial library ref number
*		pBuff		-> pointer to data to send
*		len			-> # of bytes to send
*		errP		- ptr to returned error value: 
*					0 if no error; otherwise: serErrTimeOut, serErrLineErr
*
*	CALLED BY:
*		Whoever wants to send data
*
*  RETURNS: # bytes actually sent
*
*
*************************************************************/
//DllExport
ULong SerSend(UInt refNum, VoidPtr pBuff, ULong len, Err* errP)
{
  SerGlobalsPtr	gP = PrvGlobalsP;
  DWord			commErr = 0;
  COMSTAT			comstat;
  DWord	dwErrorMask,  dwActual, dwBytesToWrite;
  Err		retval = 0;
  
  ErrFatalDisplayIf( !gP->openCount, "must open serial manager first" );
  ErrFatalDisplayIf( (refNum != PrvSavedRefNum), "invalid ref num" );
  
  
  // Validate that the address range is valid for reading.  
  //VAL_PTR_R(pBuff, len);
  
  dwErrorMask       = (DWord)0;
  dwActual          = (DWord)0;
  dwBytesToWrite     = (DWord)len;
  
  if (gP->commHdl) 
	  {
	  retval = 0;
	  
	  assert(_CrtCheckMemory());
	  
	  if (WriteFile(gP->commHdl, pBuff, dwBytesToWrite, &dwActual, NULL))
		  {
		  // When WriteFile returns TRUE(indicating success) in non-blocking mode,
		  // the operation may have still timed out, which is indicated by "actual
		  // bytes written" < "bytes to write" 
		  
		  
		  // Check for write timeout
		  if ( dwActual != dwBytesToWrite )
			  retval = serErrTimeOut;
		  }
	  else
		  {
		  commErr = GetLastError();
		  ClearCommError(gP->commHdl, &dwErrorMask, &comstat);
		  }
	  
	  }
  
  // Flush the transmit queue if we timed out
  // KK: Note: in Win32, there is no way to flush only the transmit queue
  // we have to flush both the transmit and the receive queues 
  // with FlushFileBuffers. PurgeComm() allows us to specify the
  // queue to purge but then data is deleted, not transmitted...
  if (commErr || retval == serErrTimeOut)
	  {
	  //dwPurgeFlags |= PURGE_TXCLEAR;
	  if ( ! FlushFileBuffers(gP->commHdl) )
		  {
		  commErr = GetLastError();
		  }
	  }
  // translate error value
  if (commErr)
	  {
	  if ( commErr == ERROR_COUNTER_TIMEOUT )
		  retval = serErrTimeOut;
	  else
		  retval = serErrLineErr;
	  }
  
  if (errP)
	  *errP = retval;
  
  return(dwActual);
  
}


/************************************************************
*
*  FUNCTION: SerSendWait
*
*  DESCRIPTION: Waits for all characters in the send buffer to
*		be sent over the serial port or a timeout, whichever comes
*		sooner.
*
*  PARAMETERS: 
*      refNum				-> serial library ref number
*		dwTimeOutMilliSec	-> timeout in millisec, 0 for none, -1 forever
*
*	CALLED BY:
*
*  RETURNS: 0 if no error; serErrTimeOut, serErrLineErr
*
*
*************************************************************/
#define prvFlushWaitInterval		25		// in milliseconds
Err SerSendWait(UInt refNum, Long timeOutMilliSec)
{
  SerGlobalsPtr	gP = PrvGlobalsP;
  BOOL	forever = FALSE;
  Err     err = 0;
  DWord	cbOutQueue = 0;
  DWord	dwNumIterations = 1;
  DWord   dwPurgeFlags = 0;
  DWord	commErr = 0;
  
  ErrFatalDisplayIf( !gP->openCount, "must open serial manager first" );
  ErrFatalDisplayIf( (refNum != PrvSavedRefNum), "invalid ref num" );
  
  
  // To Do - forever is dangerous - just make a big timeout for now...
  if (timeOutMilliSec == -1)
	  timeOutMilliSec = 10000;
  
  if (timeOutMilliSec == -1)
	  forever = TRUE;
  else dwNumIterations = (timeOutMilliSec / prvFlushWaitInterval) + 1;
  
  do	{
	  // data left?
	  if (PrvSerGetQueueStats(NULL, &cbOutQueue))
		  {
		  err = serErrLineErr;
		  break;
		  }
	  if (cbOutQueue == 0)
		  break;  // no bytes left
	  // try to flush buffers since there is still data left...
	  if ( ! FlushFileBuffers(gP->commHdl) )
		  {
		  commErr = GetLastError();
		  break;
		  }
	  
	  Sleep(prvFlushWaitInterval);
	  if (!forever)
		  dwNumIterations--;
	  } while ( dwNumIterations );
  
  // Timed out?
  if (PrvSerGetQueueStats(NULL, &cbOutQueue))
	  err = serErrLineErr;
  
  if (cbOutQueue != 0 && !err)
	  {
	  err = serErrTimeOut;
	  // Flush Transmit queue
	  dwPurgeFlags |= PURGE_TXCLEAR;
	  if ( !PurgeComm(gP->commHdl, dwPurgeFlags) )
		  {
		  commErr = GetLastError();
		  }
	  }
  if (commErr)
	  err = serErrLineErr;
  return(err);
  
 }


/************************************************************
*
*  FUNCTION: SerReceive
*
*  DESCRIPTION: Receives 'size' bytes worth of data or returns with
*		error if a line error or timeout is encountered
*      Note: this function counts on the Win32 ReadFile(...) to check
*      timeouts based on the SetCommTiemouts settings. (different from win16 version)
*
*  PARAMETERS: 
*		refNum		-> serial library reference number
*		pBufF		-> receive buffer
*		len			-> number of bytes to receive
*		timeout     - not used
*		errP		- ptr to error 
*
*	CALLED BY:
*		
*
*  RETURNS: # bytes received. 
*
*************************************************************/
ULong SerReceive(UInt refNum, VoidPtr pBuff, ULong len, Long timeout, 
				 Err* errP)
{
  SerGlobalsPtr	gP = PrvGlobalsP;
  COMSTAT comstat;
  DWord	dwErrorMask, dwActual, lBytesToRead;
  DWord	commErr   = 0; 
  Err		retval = serErrBadPort;
  
  ErrFatalDisplayIf( !gP->openCount, "must open serial manager first" );
  ErrFatalDisplayIf( (refNum != PrvSavedRefNum), "invalid ref num" );
  
  // Make sure that the address range is valid for read/write
  //VAL_PTR_RW(pBuff, len);
  
  
  dwActual          = 0;
  lBytesToRead      = len;
  
  if (gP->commHdl)
	  {
	  retval = 0;
	  
	  if (ReadFile(gP->commHdl, pBuff, lBytesToRead, &dwActual, NULL))
		  {
		  if (dwActual != lBytesToRead)
			  retval = serErrTimeOut;
		  }
	  else
		  {
		  commErr = GetLastError();
		  ClearCommError(gP->commHdl, &dwErrorMask, &comstat);
		  retval = serErrLineErr;
		  }
	  }
  
  
  _ASSERTE(dwActual <= len);
  
  if (errP)
	  *errP = retval;
  
  return(dwActual);
  
 }



/************************************************************
*
*  FUNCTION: SerReceiveWait
*
*  DESCRIPTION: Waits for 1 byte of data to arrive at the
*		serial port, or for a timeoout.
*
*  PARAMETERS: 
*		refNum		-> serial library reference number
*		bytes		- # bytes - not used
*						NOTE:- # bytes assumed to be 1 because that's what Win32
*						version did and SerLinkMgr always calls this with 1.
*		timeout		-> timeout in millisec. 0 for none, -1 forever
*
*	CALLED BY:
*		
*
*  RETURNS: 0 if no error; otherwise serErrTimeOut, serErrLineErr
*
*
*************************************************************/
Err SerReceiveWait(UInt refNum, ULong bytes, Long timeout)
 {
  SerGlobalsPtr	gP = PrvGlobalsP;
  Err				err = 0;
  ULong			numBytes;
  Long			startTicks;		// starting tick count


  ErrFatalDisplayIf( !gP->openCount, "must open serial manager first" );
  ErrFatalDisplayIf( (refNum != PrvSavedRefNum), "invalid ref num" );


  // wait for some data to arrive keeping track of the timeout value
  numBytes = 0;
  startTicks = GetTickCount();		// get initial tick count
  while ( TRUE)
	  {
	  // check for data in serial input buffer
	  if ((err = SerReceiveCheck( refNum, &numBytes)) != 0)
		  break;  // line error
	  else if (numBytes > 0) 
		  break;  // data
	  
	  
	  // check for timeout
	  if ( timeout != -1 && ((Long)GetTickCount() - startTicks) >= timeout )
		  {
		  err = serErrTimeOut;		// timed out
		  break;
		  }
	  }


  return( err );
 }



/************************************************************
*
*  FUNCTION: SerReceiveCheck
*
*  DESCRIPTION:Detects (does *not* read in) the precence of any 
*              incomming bytes on the comport. This routine is used as
*              a detection notifier rather that consuming any data.
*
*  PARAMETERS: 
*		refNum			- serial library reference number
*		pNumBytes       - # bytes in input queue
*
*	CALLED BY:
*		
*
*  RETURNS: 0 if no errors
*
*************************************************************/
Err SerReceiveCheck(UInt refNum, DWord* pNumBytes)
{
  SerGlobalsPtr	gP = PrvGlobalsP;
  Err       err = 0;
  
  
  ErrFatalDisplayIf( !gP->openCount, "must open serial manager first" );
  ErrFatalDisplayIf( (refNum != PrvSavedRefNum), "invalid ref num" );
  
  
  *pNumBytes = 0;
  if (PrvSerGetQueueStats(pNumBytes, NULL))
	  {
	  err = serErrLineErr;
	  }
  return err;
}			



/************************************************************
*
*  FUNCTION: SerReceiveFlush
*
*  DESCRIPTION: Flushes all characters in the receive buffer and
*		all characters coming into the serial port until no
*		more arrive within the timeout period.
*
*  PARAMETERS: 
*		refNum			-> serial library reference number
*		timeout			-- not used
*
*	CALLED BY:
*		
*
*  RETURNS: 0 if no error
*
*************************************************************/
void SerReceiveFlush(UInt refNum, Long timeout)
{
  SerGlobalsPtr	gP = PrvGlobalsP;
  DWord	dwPurgeFlags;
  
  ErrFatalDisplayIf( !gP->openCount, "must open serial manager first" );
  ErrFatalDisplayIf( (refNum != PrvSavedRefNum), "invalid ref num" );
  
  
  // KK: Note this is the way HotSync does it - the Mac serial manager
  // code reads and discards data one byte at a time
  if (gP->commHdl != NULL)
	  {
	  dwPurgeFlags = 0;
	  
	  // Receive queue
	  dwPurgeFlags |= PURGE_RXCLEAR;
	  
	  if ( !PurgeComm(gP->commHdl, dwPurgeFlags) )
		  {
		  DWord		commErr;
		  commErr = GetLastError();
		  }
	  
	  }
  return;
  
}



/************************************************************
*
*  FUNCTION: SysLibClose
*
*  DESCRIPTION: Stub so we don't have to link in SystemMgr.c
*
*  PARAMETERS: 
*			refNum - refNum of library
*
*  CALLED BY: Applications that need to uninstall a library
*
*  RETURNS: 0 if no error
*
*************************************************************/
Err	SysLibClose(UInt refNum)
{
  return( SerClose(refNum) );
}

/*************************************************************/
//	 Function:		PrvSerGetQueueStats() 
//
//	 Description:	Returns count of bytes in the input and output queues.
//
//
//	 Parameters:	pcbInQue	-- pointer to variable for returning number
//											of bytes in the input queue or NULL
//					pcbOutQue	-- pointer to variable for returning number
//											of bytes in the output queue or NULL
//
//	 Returns:		TRUE on success
//
///////////////////////////////////////////////////////////////////////////
BOOL PrvSerGetQueueStats(DWord* pcbInQue, DWord* pcbOutQue)
{
  SerGlobalsPtr	gP = PrvGlobalsP;
  BOOL		fSuccess = FALSE;
  COMSTAT	comstat;
  DWord		dwErrorMask ;
  
  
  // Make sure that the address range is
  // valid for read/write
  //if ( pcbInQue )
  //	{
  //	VAL_PTR_RW(pcbInQue, sizeof(*pcbInQue));
  //	}
  
  //if ( pcbOutQue )
  //	{
  //	VAL_PTR_RW(pcbOutQue, sizeof(*pcbOutQue));
  //	}
  
  
  if ( pcbInQue )
	  *pcbInQue = 0;
  
  if ( pcbOutQue )
	  *pcbOutQue = 0;
  
  if (gP->commHdl) 
	  {
	  fSuccess = ClearCommError(gP->commHdl, &dwErrorMask, &comstat);
	  if ( fSuccess )
		  {
		  if ( pcbInQue )
			  *pcbInQue = comstat.cbInQue;
		  
		  if ( pcbOutQue )
			  *pcbOutQue = comstat.cbOutQue;
		  }
	  }
  
  return(!fSuccess);
}


