/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"

#include "CPU_REG.h"			// Screen::InvalidateAll, Hardware::PenEvent, etc.

#include <FL/Fl.H>
#include <FL/fl_draw.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Pixmap.H>
#include <FL/Fl_Window.H>
#include <FL/math.h>

#include "fltk_MainWindow.h"
#include "fltk_LcdWindow.h"
#include "MapFile.h"
#include "Preferences.h"

#include "POSE.xpm"

const float REDRAW_FREQ = 0.05;   // = 20x / second

// X, Y, W, H
#define POWER_RECT 0,276,17,23
#define DATEBOOK_RECT 26,276,28,23
#define PHONE_RECT 64,276,28,23
#define TODO_RECT 135,276,28,23
#define MEMO_RECT 172,276,28,23
#define SCROLL_UP_RECT 99,275,27,12
#define SCROLL_DN_RECT 99,296,27,12

// -----------------------------------------------------------------------------
// file static methods
// -----------------------------------------------------------------------------

char **loadXPMFile( const char* fname );
void freeXPMData( char **&data );

// -----------------------------------------------------------------------------
// Fl_Pixmap helper class
// -----------------------------------------------------------------------------

// a Fl_Pixmap subclass which owns dynamicly allocated pixmap data
class Dynamic_Pixmap : public Fl_Pixmap
{
public:
    Dynamic_Pixmap( char **data, bool dyn=true ) : Fl_Pixmap( data ), _data( data ), _dyn(dyn) {}
    virtual ~Dynamic_Pixmap( void ) { if ( _dyn && (_data!=NULL) ) freeXPMData( _data ); }
    virtual void measure( int &w, int &h );
private:
    char **_data;
    bool _dyn;
};

void Dynamic_Pixmap::measure( int &w, int &h )
{
    int ncolors, chars_per_pixel;
    sscanf(data[0],"%d %d %d %d",&w,&h,&ncolors,&chars_per_pixel);
}

// -----------------------------------------------------------------------------
// redraw the LCD every REDRAW_FREQ seconds
static void damageLcd( void *data )
{
    LCD_window *lcd = (LCD_window*) data;
    if ( lcd->isOn() )
    {
        lcd->damageLcdArea();
        Fl::add_timeout( REDRAW_FREQ, damageLcd, lcd );
    }
}

// -----------------------------------------------------------------------------
// constructor / destructor
// -----------------------------------------------------------------------------

LCD_window::LCD_window( MainWindow *parent )
        : Fl_Double_Window( 0, 0, 600, 600, "" ),
          _parent(parent), _doc( NULL )
{
    _buffer = NULL;
    _lcdOn = true;
    _case = NULL;
    
    resetToDefault();
    initialize();
    
    Fl::add_timeout( REDRAW_FREQ, damageLcd, this );
}

LCD_window::~LCD_window( void )
{
    _bufferLock.lock();
    if ( _buffer != NULL )
        delete _buffer;
    _bufferLock.unlock();
}

// -----------------------------------------------------------------------------
// public methods
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// Fl_Double_Window overrides
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// Handle the GUI event. This routine translates the FLTK GUI events into
// the corresponding emulator actions.
int LCD_window::handle( int event )
{
    bool inTouchScreen = Fl::event_inside( _lcdX, _lcdY, lcdWidth * _scale, (lcdHeight + silkScreenHeight)* _scale );
    int ex = Fl::event_x();
    int ey = Fl::event_y();

    int lcdx = (ex - _lcdX) / _scale;
    int lcdy = (ey - _lcdY) / _scale;

    switch (event) {
        case FL_PUSH:
        case FL_DRAG:
            if ( Fl::event_button() == 3 )
            {
                _parent->popupMenu();
                return 1;
            }
            if ( _penTracking )
            {
                bound( lcdx, lcdy, 0, 0, lcdWidth, lcdHeight + silkScreenHeight );
                if ( _doc != NULL )
                    _doc->pen( true, lcdx, lcdy );
                return 1;
            }
            else if ( inTouchScreen )
            {
                _penTracking = true;
                if ( _doc != NULL )
                    _doc->pen( true, lcdx, lcdy );
                return 1;
            }
            else if ( _currentButton == kButton_None )
            {
                if ( ptInRect( ex, ey, POWER_RECT ) )
                {
                    _currentButton = kButton_Power;
                }
                else if ( ptInRect( ex, ey, DATEBOOK_RECT ) )
                {
                    _currentButton = kButton_DateBook;
                }
                else if ( ptInRect( ex, ey, PHONE_RECT ) )
                {
                    _currentButton = kButton_Phone;
                }
                else if ( ptInRect( ex, ey, TODO_RECT ) )
                {
                    _currentButton = kButton_ToDo;
                }
                else if ( ptInRect( ex, ey, MEMO_RECT ) )
                {
                    _currentButton = kButton_Memo;
                }
                else if ( ptInRect( ex, ey, SCROLL_UP_RECT ) )
                {
                    _currentButton = kButton_UpArrow;
                }
                else if ( ptInRect( ex, ey, SCROLL_DN_RECT ) )
                {
                    _currentButton = kButton_DownArrow;
                }
                if ( _currentButton != kButton_None )
                {
                    if ( _doc != NULL )
                        _doc->button( true, _currentButton );
                    return 1;
                }
            }
        case FL_RELEASE:
            if ( _penTracking )
            {
                bound( lcdx, lcdy, 0, 0, lcdWidth, lcdHeight + silkScreenHeight );
                _penTracking = false;
                if ( _doc != NULL )
                    _doc->pen( false, lcdx, lcdy );
                return 1;
            }
            else if ( _currentButton != kButton_None )
            {
                if ( _doc != NULL )
                    _doc->button( false, _currentButton );
                _currentButton = kButton_None;
                return 1;
            }

        case FL_FOCUS:
            // always accept the focus
            return 1;

        case FL_KEYBOARD:
        {
            if (Fl::event_state(FL_ALT|FL_META))
            {
                // reserved for shortcuts
                return 0;
            }

            int c = Fl::event_key();

            // handle prinable characters
            if ( strlen( Fl::event_text() ) > 0 )
            {
                if ( _doc != NULL )
                    _doc->putkey( Fl::event_text()[0] );
                return 1;
            }

            // handle all other characters
            switch ( c )
            {
                case FL_Enter:
                case FL_KP_Enter:
                    if ( _doc != NULL )
                        _doc->putkey( '\n' );
                    return 1;

                default:
                    return 0;
            }
        }

        default:
            return Fl_Double_Window::handle(event);
    }
}

// -----------------------------------------------------------------------------
// draw the case and the inner LCD screen
void LCD_window::draw()
{
    // redraw everything if anything other than just the lcd damage bit is on
    bool drawAll = ( damage() & ~LCD_DAMAGE_BIT ) != 0;

    fl_color( FL_WHITE );
    if ( drawAll )
    {
        _case->draw( 0,0 );
    }

    // only draw the LCD if we have a document & the LCD is on
    if ( _lcdOn && (_doc != NULL) )
    {
        _bufferLock.lock();
        if ( updateBuffer( drawAll ) || drawAll )
        {
            int w = lcdWidth * _scale;
            int h = lcdHeight * _scale;
            fl_draw_image( _buffer, _lcdX, _lcdY, w, h );
        }
        _bufferLock.unlock();
    }

}

// -----------------------------------------------------------------------------
// turn the lcd off or on
void LCD_window::setOn( bool on )
{
    if ( !_lcdOn && on )
    {
        // turn on lcd
        initialize();
        _lcdOn = true;
        Fl::add_timeout( REDRAW_FREQ, damageLcd, this );
        redraw();
    }
    else if ( _lcdOn && !on )
    {
        // turn lcd off
        _lcdOn = false;
        _bufferLock.lock();
        if ( _buffer != NULL )
        {
            delete _buffer;
            _buffer = NULL;
        }
        _bufferLock.unlock();
        redraw();
    }
}

// -----------------------------------------------------------------------------
// reset configurable parameters to default values
void LCD_window::resetToDefault()
{
    _currentButton = kButton_None;
    _penTracking = false;

    if ( _case != NULL )
        delete _case;
    _case = new Dynamic_Pixmap( POSE, false );

    _lcdX = 32;
    _lcdY = 32;
    _scale = 1;
    _blRGB = 0x0064F0DC; // nice glowing blue
    _bgRGB = 0x00FFFFFF; // white
    _useBGColor = false;
}

// -----------------------------------------------------------------------------
// initialize data & presentation
void LCD_window::initialize()
{
    _bufferLock.lock();
    if ( _buffer != NULL )
    {
        delete _buffer;
        _buffer = NULL;
    }
    _buffer = new unsigned char[ lcdWidth * _scale * lcdHeight * _scale * 3 ];
    _bufferLock.unlock();

    // resize this window to the size of the surrounding case
    int w, h;
    _case->measure( w, h );
    size( w, h );
    _parent->size(w,h);
}

// -----------------------------------------------------------------------------
// load the images and data specified in the given skin file...
void LCD_window::loadSkinFile( const char* fname )
{
    // first reset to default values, in case skin file doesn't set some parameters
    resetToDefault();

    if ( strcmp( fname, BASE_SKIN_NAME ) == 0 )
    {
        initialize();
        redraw();
        return;
    }
    
    // load and parse skin file
    MapFile m( fname );

    char casefile[PATH_MAX];
    char *f = m.find( "Case" );
    if ( f != NULL )
    {
        if ( *f == '/' )
        {
            // absolute path
            strcpy( casefile, f );
        }
        else
        {
            // relative path
            char *p = strrchr( fname , '/' );
            int len = p - fname + 1;
            strncpy( casefile, fname, len );
            strcpy( &casefile[len], f );
        }

        char **casedata = loadXPMFile( casefile );
        if ( casedata != NULL )
        {
            if ( _case != NULL )
                delete _case;
            _case = new Dynamic_Pixmap( casedata );
        }
    }
    
    // read in the lcd x and y offsets and scale
    char *buf = m.find( "LcdX" );
    if ( buf != NULL )
    {
        sscanf( buf, "%d", &_lcdX );
    }
    buf = m.find( "LcdY" );
    if ( buf != NULL )
    {
        sscanf( buf, "%d", &_lcdY );
    }
    buf = m.find( "Scale" );
    if ( buf != NULL )
    {
        sscanf( buf, "%d", &_scale );
    }
    buf = m.find( "BacklightColor" );
    if ( buf != NULL )
    {
        sscanf( buf, "%lx", &_blRGB );
    }
    buf = m.find( "BackgroundColor" );
    if ( buf != NULL )
    {
        if ( sscanf( buf, "%lx", &_bgRGB ) == 1 )
        {
            _useBGColor = true;
        }
    }

    initialize();
    
    // force a redraw with this new theme
    redraw();
}

// -----------------------------------------------------------------------------
// private methods
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// update the offscreen lcd buffer from the emulator's screen RAM. Returns true
// if the offscreen buffer data has been changed, false otherwise. Note that
// you must own the buffer lock to call this method.
int LCD_window::updateBuffer( int always )
{
    if (always)
    {
        Screen::InvalidateAll ();
    }

    // Get information about the screen.  Return immediately if nothing's changed.
    Screen::BufferInfo	info;
    if (!Screen::GetBits (info, true /* incremental */))
        return false;

    // Get the bounds of the area we'll be blitting from.
    int x = info.visibleBounds.left;
    int y = info.firstLine;
    int width = info.visibleBounds.right - x;
    int height = info.lastLine - y;

    // There is a moment during boot-up when the LCD width is 1024.
    // We don't really want that as the width (we only allow for 160
    // pixels in our window), so clip to 160.
    if (width > lcdWidth)
    {
        width = lcdWidth;
    }

    if ( info.myBuffer == NULL )
    {
        // during a reset, screen buffer may be empty
        return false;
    }

    // buffer access macros
#define SRCPTR( r, c ) ((unsigned char*) info.myBuffer) + (r) * info.myRowBytes + (c)
#define DESTPTR( r, c, n ) _buffer + (_scale * (lcdWidth*(r*_scale + n) + c) * 3)
    
// CLUT access macros
#define CLUT_R(x) info.myCLUT[(x)].r
#define CLUT_G(x) info.myCLUT[(x)].g
#define CLUT_B(x) info.myCLUT[(x)].b
    
    // save the current BG color in case we have to change it.
    Screen::RGBType saveBG = info.myCLUT[0];
    
    bool switchBGColor = false;


    // change the background color if backlight is on
    if ( info.backlightOn )
    {
        CLUT_R(0) = (unsigned char)((0x00FF0000 & _blRGB)>>16);
        CLUT_G(0) = (unsigned char)((0x0000FF00 & _blRGB)>>8);
        CLUT_B(0) = (unsigned char)(0x000000FF & _blRGB);
        switchBGColor = true;
    }
    else if ( _useBGColor )
    {
        // if the skin specifies a background color, then use that
        CLUT_R(0) = (unsigned char)((0x00FF0000 & _bgRGB)>>16);
        CLUT_G(0) = (unsigned char)((0x0000FF00 & _bgRGB)>>8);
        CLUT_B(0) = (unsigned char)(0x000000FF & _bgRGB);
        switchBGColor = true;
    }
    

    // convert 8bpp Screen RAM data to 24bpp buffer data
    switch ( _scale )
    {
        default:
        case 1:
        {
            for (int ly=y; ly<y+height; ly++)
                for (int lx=x; lx<x+width; lx++)
                {
                    unsigned char pix = *(SRCPTR(ly,lx));
                    unsigned char *p = DESTPTR(ly,lx,0);
                    *p++=CLUT_R(pix); *p++=CLUT_G(pix); *p=CLUT_B(pix);
                }
            break;
        }
        case 2:
        {
            for (int ly=y; ly<y+height; ly++)
                for (int lx=x; lx<x+width; lx++)
                {
                    unsigned char pix = *(SRCPTR(ly,lx));
                    unsigned char *p = DESTPTR(ly,lx,0);
                    *p++=CLUT_R(pix); *p++=CLUT_G(pix); *p++=CLUT_B(pix);
                    *p++=CLUT_R(pix); *p++=CLUT_G(pix); *p=CLUT_B(pix);
                    p = DESTPTR(ly,lx,1);
                    *p++=CLUT_R(pix); *p++=CLUT_G(pix); *p++=CLUT_B(pix);
                    *p++=CLUT_R(pix); *p++=CLUT_G(pix); *p=CLUT_B(pix);
                }
            break;
        }
    }


    if ( switchBGColor )
    {
        info.myCLUT[0] = saveBG;
    }



    return true;
}

// -----------------------------------------------------------------------------
// bound the given point (x,y) within the given box
void LCD_window::bound( int& x, int& y, int bx, int by, int bw, int bh )
{
    if ( x < bx ) x = bx;
    if ( x >= bx + bw ) x = bx + bw - 1;
    if ( y < by ) y = by;
    if ( y >= by + bh ) y = by + bh - 1;
}

// -----------------------------------------------------------------------------
// pt in rect
bool LCD_window::ptInRect( int x, int y, int rx, int ry, int rw, int rh )
{
    if ( x >= rx && x < (rx + rw) && y >=ry && y < (ry + rh) )
        return true;
    return false;
}

// -----------------------------------------------------------------------------
// load a pixmap from a .xpm file
const unsigned MAXLINE = 1024;
char **allocateXPMData( int lines )
{
    char **result = new char*[lines];
    for (int i=0; i<lines; ++i)
    {
        result[i] = NULL;
    }
    return result;
}

void freeXPMData( char **&data )
{
    if ( data[0] != NULL )
    {
        unsigned int w,h,cc,bps;
        char buffer[ MAXLINE ];
        if ( sscanf( buffer, "%d %d %d %d", &w, &h, &cc, &bps ) == 4 )
        {
            int lines = 1 + cc + h;
            for (int i=0; i<lines; ++i)
            {
                if ( data[i] != NULL )
                    delete data[i];
            }
        }
    }
    delete data;
}

char **loadXPMFile( const char* fname )
{
    FILE *f=fopen( fname , "rb" );
    
    if ( f == NULL )
    {
        return NULL;
    }
    
    char buffer[ MAXLINE ];
    char *ptr = buffer;
    char **data = NULL;
    char c, last = '\0';
    bool incomment = false;
    bool indata = false;
    bool gotmetrics = false;
    int count = 0;
    
    while ( (c = fgetc( f )) != EOF )
    {
        switch ( c )
        {
            case '*':
                if ( last == '/' )
                    incomment = true;
                else if (indata)
                    *ptr++ = c;
                break;
            case '/':
                if ( incomment && (last == '*' ) )
                    incomment = false;
                else if (indata)
                    *ptr++ = c;
                break;
            case '"':
                if ( indata )
                {
                    *ptr = '\0';
                    if ( !gotmetrics )
                    {
                        unsigned int w,h,cc,bps;
                        if ( sscanf( buffer, "%d %d %d %d", &w, &h, &cc, &bps ) != 4 )
                        {
                            return NULL;
                        }
                        data = allocateXPMData( 1 + cc + h );
                        gotmetrics = true;
                    }
                    data[count] = new char[ strlen(buffer) + 1 ];
                    strcpy( data[count++], buffer );
                    ptr = buffer;
                    indata = false;
                }
                else
                    indata = true;
                break;
            default:
                if (indata) *ptr++ = c;
                break;
        }
        last = c;
    }

    fclose( f );
    return data;
}
